/*
 * Clock and power management. This exposes the combined functionality
 * provided by the X1600 and related SoCs. The power management
 * functionality could be exposed using a separate driver.
 *
 * Copyright (C) 2017, 2018, 2020, 2021, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include "cpm-x1600.h"



// Register locations.

enum Regs : unsigned
{
  Clock_control       = 0x000,  // CPCCR
  Low_power_control   = 0x004,  // LCR
  Clock_gate0         = 0x020,  // CLKGR0
  Clock_gate1         = 0x028,  // CLKGR1
  Sleep_control       = 0x024,  // OPCR (oscillator and power control)
  Clock_status        = 0x0d4,  // CPCSR

  Divider_can0        = 0x0a0,  // CAN0CDR
  Divider_can1        = 0x0a8,  // CAN1CDR
  Divider_cdbus       = 0x0ac,  // CDBUSCDR
  Divider_cim         = 0x078,  // CIMCDR
  Divider_ddr         = 0x02c,  // DDRCDR
  Divider_mac         = 0x054,  // MACCDR
  Divider0_i2s0       = 0x060,  // I2S0CDR
  Divider1_i2s0       = 0x070,  // I2S0CDR1
  Divider0_i2s1       = 0x07c,  // I2S1CDR (from X2000 manual)
  Divider1_i2s1       = 0x080,  // I2S1CDR1 (from X2000 manual)
  Divider_lcd         = 0x064,  // LPCDR
  Divider_macphy0     = 0x0e4,  // MPHY0C
  Divider_msc0        = 0x068,  // MSC0CDR
  Divider_msc1        = 0x0a4,  // MSC1CDR
  Divider_pwm         = 0x06c,  // PWMCDR
  Divider_sfc         = 0x074,  // SFCCDR
  Divider_ssi         = 0x05c,  // SSICDR

  Cpm_interrupt       = 0x0b0,  // CPM_INTR
  Cpm_interrupt_en    = 0x0b4,  // CPM_INTRE
  Cpm_swi             = 0x0bc,  // CPM_SFTINT
  Cpm_scratch         = 0x034,  // CPSPR
  Cpm_scratch_prot    = 0x038,  // CPSPPR

  Gate_ddr            = 0x0d0,  // DRCG

  Usb_param_control0  = 0x03c,  // USBPCR
  Usb_reset_detect    = 0x040,  // USBRDT
  Usb_vbus_jitter     = 0x044,  // USBVBFIL
  Usb_param_control1  = 0x048,  // USBPCR1

  Pll_control         = 0x00c,  // CPPCR
  Pll_control_A       = 0x010,  // CPAPCR
  Pll_control_M       = 0x014,  // CPMPCR
  Pll_control_E       = 0x018,  // CPEPCR
  Pll_fraction_A      = 0x084,  // CPAPACR
  Pll_fraction_M      = 0x088,  // CPMPACR
  Pll_fraction_E      = 0x08c,  // CPEPACR
};



// Register field definitions.

static Field Clock_source_main         (Clock_control, 3, 30),       // SEL_SRC (output to SCLK_A)
             Clock_source_cpu          (Clock_control, 3, 28),       // SEL_CPLL (output to CCLK)
             Clock_source_hclock0      (Clock_control, 3, 26),       // SEL_H0PLL (output to AHB0)
             Clock_source_hclock2      (Clock_control, 3, 24),       // SEL_H2PLL (output to AHB2)
             Clock_source_can0         (Divider_can0, 3, 30),        // CA0CS
             Clock_source_can1         (Divider_can1, 3, 30),        // CA1CS
             Clock_source_cdbus        (Divider_cdbus, 3, 30),       // CDCS
             Clock_source_cim          (Divider_cim, 3, 30),         // CIMPCS
             Clock_source_ddr          (Divider_ddr, 3, 30),         // DCS
             Clock_source_i2s0         (Divider0_i2s0, 1, 30),       // I2PCS
             Clock_source_i2s1         (Divider0_i2s1, 1, 30),       // I2PCS
             Clock_source_lcd          (Divider_lcd, 3, 30),         // LPCS
             Clock_source_mac          (Divider_mac, 3, 30),         // MACPCS
             Clock_source_msc0         (Divider_msc0, 3, 30),        // MPCS
             Clock_source_msc1         (Divider_msc1, 3, 30),        // MPCS
             Clock_source_pwm          (Divider_pwm, 3, 30),         // PWMPCS
             Clock_source_sfc          (Divider_sfc, 3, 30),         // SFCS
             Clock_source_ssi          (Divider_ssi, 3, 30),         // SPCS

             Clock_busy_cpu            (Clock_status, 1, 0),
             Clock_busy_hclock0        (Clock_status, 1, 1),
             Clock_busy_hclock2        (Clock_status, 1, 2),
             Clock_busy_ddr            (Divider_ddr, 1, 28),
             Clock_busy_mac            (Divider_mac, 1, 28),
             Clock_busy_lcd            (Divider_lcd, 1, 28),
             Clock_busy_msc0           (Divider_msc0, 1, 28),
             Clock_busy_msc1           (Divider_msc1, 1, 28),
             Clock_busy_sfc            (Divider_sfc, 1, 28),
             Clock_busy_ssi            (Divider_ssi, 1, 28),
             Clock_busy_cim            (Divider_cim, 1, 28),
             Clock_busy_pwm            (Divider_pwm, 1, 28),
             Clock_busy_can0           (Divider_can0, 1, 28),
             Clock_busy_can1           (Divider_can1, 1, 28),
             Clock_busy_cdbus          (Divider_cdbus, 1, 28),

             Clock_change_enable_cpu   (Clock_control, 1, 22),
             Clock_change_enable_ahb0  (Clock_control, 1, 21),
             Clock_change_enable_ahb2  (Clock_control, 1, 20),
             Clock_change_enable_ddr   (Divider_ddr, 1, 29),
             Clock_change_enable_mac   (Divider_mac, 1, 29),
             Clock_gate_i2s0           (Divider0_i2s0, 1, 29),       // CE_I2S is gate, not change enable
             Clock_gate_i2s1           (Divider0_i2s1, 1, 29),       // CE_I2S is gate, not change enable
             Clock_change_enable_lcd   (Divider_lcd, 1, 29),
             Clock_change_enable_msc0  (Divider_msc0, 1, 29),
             Clock_change_enable_msc1  (Divider_msc1, 1, 29),
             Clock_change_enable_sfc   (Divider_sfc, 1, 29),
             Clock_change_enable_ssi   (Divider_ssi, 1, 29),
             Clock_change_enable_cim   (Divider_cim, 1, 29),
             Clock_change_enable_pwm   (Divider_pwm, 1, 29),
             Clock_change_enable_can0  (Divider_can0, 1, 29),
             Clock_change_enable_can1  (Divider_can1, 1, 29),
             Clock_change_enable_cdbus (Divider_cdbus, 1, 29),

             Clock_divider_cpu         (Clock_control, 0x0f, 0),     // CDIV
             Clock_divider_hclock0     (Clock_control, 0x0f, 8),     // H0DIV (fast AHB peripherals)
             Clock_divider_hclock2     (Clock_control, 0x0f, 12),    // H2DIV (fast AHB peripherals)
             Clock_divider_l2cache     (Clock_control, 0x0f, 4),     // L2CDIV
             Clock_divider_pclock      (Clock_control, 0x0f, 16),    // PDIV (slow APB peripherals)
             Clock_divider_can0        (Divider_can0, 0xff, 0),      // CAN0CDR
             Clock_divider_can1        (Divider_can1, 0xff, 0),      // CAN1CDR
             Clock_divider_cdbus       (Divider_cdbus, 0xff, 0),     // CDBUSCDR
             Clock_divider_cim         (Divider_cim, 0xff, 0),       // CIMCDR
             Clock_divider_ddr         (Divider_ddr, 0x0f, 0),       // DDRCDR
             Clock_divider_i2s0_m      (Divider0_i2s0, 0x1ff, 20),   // I2SDIV_M
             Clock_divider_i2s0_n      (Divider0_i2s0, 0xfffff, 0),  // I2SDIV_N
             Clock_divider_i2s0_d      (Divider1_i2s0, 0xfffff, 0),  // I2SDIV_D
             Clock_divider_i2s1_m      (Divider0_i2s1, 0x1ff, 20),   // I2SDIV_M
             Clock_divider_i2s1_n      (Divider0_i2s1, 0xfffff, 0),  // I2SDIV_N
             Clock_divider_i2s1_d      (Divider1_i2s1, 0xfffff, 0),  // I2SDIV_D
             Clock_divider_lcd         (Divider_lcd, 0xff, 0),       // LPCDR
             Clock_divider_mac         (Divider_mac, 0xff, 0),       // MACCDR
             Clock_divider_msc0        (Divider_msc0, 0xff, 0),      // MSC0CDR
             Clock_divider_msc1        (Divider_msc1, 0xff, 0),      // MSC1CDR
             Clock_divider_pwm         (Divider_pwm, 0x0f, 0),       // PWMCDR
             Clock_divider_sfc         (Divider_sfc, 0xff, 0),       // SFCCDR
             Clock_divider_ssi         (Divider_ssi, 0xff, 0),       // SSICDR

             Clock_divider_i2s0_n_auto (Divider1_i2s0, 1, 31),       // I2S_NEN
             Clock_divider_i2s0_d_auto (Divider1_i2s0, 1, 30),       // I2S_DEN
             Clock_divider_i2s1_n_auto (Divider1_i2s1, 1, 31),       // I2S_NEN
             Clock_divider_i2s1_d_auto (Divider1_i2s1, 1, 30),       // I2S_DEN

             Clock_gate_main           (Clock_control, 1, 23, true), // GATE_SCLKA
             Clock_gate_ddr            (Clock_gate0, 1, 31, true),   // DDR
             Clock_gate_ahb0           (Clock_gate0, 1, 29, true),   // AHB0
             Clock_gate_apb0           (Clock_gate0, 1, 28, true),   // APB0
             Clock_gate_rtc            (Clock_gate0, 1, 27, true),   // RTC
             Clock_gate_aes            (Clock_gate0, 1, 24, true),   // AES
             Clock_gate_lcd_pixel      (Clock_gate0, 1, 23, true),   // LCD
             Clock_gate_cim            (Clock_gate0, 1, 22, true),   // CIM
             Clock_gate_dma            (Clock_gate0, 1, 21, true),   // PDMA
             Clock_gate_ost            (Clock_gate0, 1, 20, true),   // OST
             Clock_gate_ssi0           (Clock_gate0, 1, 19, true),   // SSI0
             Clock_gate_timer          (Clock_gate0, 1, 18, true),   // TCU
             Clock_gate_dtrng          (Clock_gate0, 1, 17, true),   // DTRNG
             Clock_gate_uart2          (Clock_gate0, 1, 16, true),   // UART2
             Clock_gate_uart1          (Clock_gate0, 1, 15, true),   // UART1
             Clock_gate_uart0          (Clock_gate0, 1, 14, true),   // UART0
             Clock_gate_sadc           (Clock_gate0, 1, 13, true),   // SADC
             Clock_gate_aic            (Clock_gate0, 1, 11, true),   // AUDIO
             Clock_gate_ssi_slv        (Clock_gate0, 1, 10, true),   // SSI_SLV
             Clock_gate_i2c1           (Clock_gate0, 1, 8, true),    // I2C1
             Clock_gate_i2c0           (Clock_gate0, 1, 7, true),    // I2C0
             Clock_gate_msc1           (Clock_gate0, 1, 5, true),    // MSC1
             Clock_gate_msc0           (Clock_gate0, 1, 4, true),    // MSC0
             Clock_gate_otg            (Clock_gate0, 1, 3, true),    // OTG
             Clock_gate_sfc            (Clock_gate0, 1, 2, true),    // SFC
             Clock_gate_efuse          (Clock_gate0, 1, 1, true),    // EFUSE
             Clock_gate_nemc           (Clock_gate0, 1, 0, true),    // NEMC
             Clock_gate_arb            (Clock_gate1, 1, 30, true),   // ARB
             Clock_gate_mipi_csi       (Clock_gate1, 1, 28, true),   // MIPI_CSI
             Clock_gate_intc           (Clock_gate1, 1, 26, true),   // INTC
             Clock_gate_gmac0          (Clock_gate1, 1, 23, true),   // GMAC0
             Clock_gate_uart3          (Clock_gate1, 1, 16, true),   // UART3
             Clock_gate_i2s0_tx        (Clock_gate1, 1, 9, true),    // I2S0_dev_tclk
             Clock_gate_i2s0_rx        (Clock_gate1, 1, 8, true),    // I2S0_dev_rclk
             Clock_gate_hash           (Clock_gate1, 1, 6, true),    // HASH
             Clock_gate_pwm            (Clock_gate1, 1, 5, true),    // PWM
             Clock_gate_cdbus          (Clock_gate1, 1, 2, true),    // CDBUS
             Clock_gate_can1           (Clock_gate1, 1, 1, true),    // CAN1
             Clock_gate_can0           (Clock_gate1, 1, 0, true),    // CAN0

             Pll_enable_A              (Pll_control_A, 1, 0),        // APLLEN
             Pll_enable_E              (Pll_control_E, 1, 0),        // EPLLEN
             Pll_enable_M              (Pll_control_M, 1, 0),        // MPLLEN

             Pll_stable_A              (Pll_control_A, 1, 3),        // APLL_ON
             Pll_stable_E              (Pll_control_E, 1, 3),        // EPLL_ON
             Pll_stable_M              (Pll_control_M, 1, 3),        // MPLL_ON

             Pll_bypass_A              (Pll_control_A, 1, 30),       // APLL_BP
             Pll_bypass_E              (Pll_control_E, 1, 26),       // EPLL_BP
             Pll_bypass_M              (Pll_control_M, 1, 28),       // MPLL_BP

             Pll_multiplier_A          (Pll_control_A, 0xfff, 20),   // APLLM
             Pll_multiplier_E          (Pll_control_E, 0x3f, 20),    // EPLLM (observed)
             Pll_multiplier_M          (Pll_control_M, 0xfff, 20),   // MPLLM

             Pll_input_division_A      (Pll_control_A, 0x3f, 14),    // APLLN
             Pll_input_division_E      (Pll_control_E, 0x3f, 14),    // EPLLN
             Pll_input_division_M      (Pll_control_M, 0x3f, 14),    // MPLLN

             Pll_output_division1_A    (Pll_control_A, 0x07, 11),    // APLLOD1
             Pll_output_division1_E    (Pll_control_E, 0x07, 11),    // EPLLOD1
             Pll_output_division1_M    (Pll_control_M, 0x07, 11),    // MPLLOD1

             Pll_output_division0_A    (Pll_control_A, 0x07, 8),     // APLLOD0
             Pll_output_division0_E    (Pll_control_E, 0x07, 8),     // EPLLOD0
             Pll_output_division0_M    (Pll_control_M, 0x07, 8);     // MPLLOD0



// Multiplexer instances.

#define Clocks(...) ((enum Clock_identifiers []) {__VA_ARGS__})

static Mux mux_external (Clock_external),

           // Clocks being propagated to others.

           mux_hclock0        (Clock_hclock0),
           mux_hclock2        (Clock_hclock2),
           mux_pclock         (Clock_pclock),
           mux_hclock2_pclock (Clock_hclock2_pclock),
           mux_i2s0_rx        (Clock_i2s0),
           mux_i2s0_tx        (Clock_i2s1),

           // Main peripheral and bus clock sources.

           mux_bus      (4, Clocks(Clock_main, Clock_pll_M, Clock_pll_E, Clock_external)),
           mux_core     (3, Clocks(Clock_none, Clock_main, Clock_pll_M)),
           mux_dev      (3, Clocks(Clock_main, Clock_pll_M, Clock_pll_E)),
           mux_main     (3, Clocks(Clock_none, Clock_external, Clock_pll_A)),
           mux_i2s      (2, Clocks(Clock_main, Clock_pll_E));



// Clock instances.

static Clock_null        clock_none;

static Clock_passive     clock_external(24000000);

// Note the use of extra parentheses due to the annoying C++ "most vexing parse"
// problem. See: https://en.wikipedia.org/wiki/Most_vexing_parse

static Clock clock_aic((Source(mux_hclock2)), (Control(Clock_gate_aic))),

             clock_dma((Source(mux_hclock2)), (Control(Clock_gate_dma))),

             clock_i2c0((Source(mux_pclock)), (Control(Clock_gate_i2c0))),

             clock_i2c1((Source(mux_pclock)), (Control(Clock_gate_i2c1))),

             clock_i2s0(Source(mux_i2s, Clock_source_i2s0), Control(Clock_gate_i2s0)),

             clock_i2s1(Source(mux_i2s, Clock_source_i2s1), Control(Clock_gate_i2s1)),

             clock_main(Source(mux_main, Clock_source_main), Control(Clock_gate_main)),

             clock_mipi_csi((Source(mux_hclock0)), Control(Clock_gate_mipi_csi)),

             clock_otg0((Source(mux_hclock2)), (Control(Clock_gate_otg))),

             clock_timer((Source(mux_pclock)), (Control(Clock_gate_timer))),

             clock_uart0((Source(mux_external)), (Control(Clock_gate_uart0))),

             clock_uart1((Source(mux_external)), (Control(Clock_gate_uart1))),

             clock_uart2((Source(mux_external)), (Control(Clock_gate_uart2))),

             clock_uart3((Source(mux_external)), (Control(Clock_gate_uart3))),

             // Special parent clock for hclock2 and pclock.

             clock_hclock2_pclock(Source(mux_core, Clock_source_hclock2),
                           Control(Clock_gate_apb0, Clock_change_enable_ahb2));

static Clock_divided
             clock_can0(Source(mux_bus, Clock_source_can0),
                        Control(Clock_gate_can0, Clock_change_enable_can0, Clock_busy_can0),
                        Divider(Clock_divider_can0)),

             clock_can1(Source(mux_bus, Clock_source_can1),
                        Control(Clock_gate_can1, Clock_change_enable_can1, Clock_busy_can1),
                        Divider(Clock_divider_can1)),

             clock_cdbus(Source(mux_dev, Clock_source_cdbus),
                         Control(Clock_gate_cdbus, Clock_change_enable_cdbus, Clock_busy_cdbus),
                         Divider(Clock_divider_cdbus)),

             clock_cim(Source(mux_dev, Clock_source_cim),
                       Control(Clock_gate_cim, Clock_change_enable_cim, Clock_busy_cim),
                       Divider(Clock_divider_cim)),

             clock_cpu(Source(mux_core, Clock_source_cpu),
                       Control(Field::undefined, Clock_change_enable_cpu, Clock_busy_cpu),
                       Divider(Clock_divider_cpu)),

             clock_ddr(Source(mux_core, Clock_source_ddr),
                       Control(Clock_gate_ddr, Clock_change_enable_ddr, Clock_busy_ddr),
                       Divider(Clock_divider_ddr)),

             clock_hclock0(Source(mux_core, Clock_source_hclock0),
                           Control(Clock_gate_ahb0, Clock_change_enable_ahb0),
                           Divider(Clock_divider_hclock0)),

             clock_hclock2((Source(mux_hclock2_pclock)), (Divider(Clock_divider_hclock2))),

             clock_lcd_pixel(Source(mux_dev, Clock_source_lcd),
                             Control(Clock_gate_lcd_pixel, Clock_change_enable_lcd, Clock_busy_lcd),
                             Divider(Clock_divider_lcd)),

             clock_mac(Source(mux_dev, Clock_source_mac),
                       Control(Clock_gate_gmac0, Clock_change_enable_mac, Clock_busy_mac),
                       Divider(Clock_divider_mac)),

             clock_msc0(Source(mux_dev, Clock_source_msc0),
                        Control(Clock_gate_msc0, Clock_change_enable_msc0, Clock_busy_msc0),
                        Divider(Clock_divider_msc0)),

             clock_msc1(Source(mux_dev, Clock_source_msc1),
                        Control(Clock_gate_msc1, Clock_change_enable_msc1, Clock_busy_msc1),
                        Divider(Clock_divider_msc1)),

             clock_pclock((Source(mux_hclock2_pclock)), (Divider(Clock_divider_pclock))),

             clock_pwm0(Source(mux_dev, Clock_source_pwm),
                        Control(Clock_gate_pwm, Clock_change_enable_pwm, Clock_busy_pwm),
                        Divider(Clock_divider_pwm)),

             clock_sfc(Source(mux_dev, Clock_source_sfc),
                       Control(Clock_gate_sfc, Clock_change_enable_sfc, Clock_busy_sfc),
                       Divider(Clock_divider_sfc)),

             clock_ssi0(Source(mux_dev, Clock_source_ssi),
                        Control(Clock_gate_ssi0, Clock_change_enable_ssi, Clock_busy_ssi),
                        Divider(Clock_divider_ssi));

static Clock_divided_i2s
             clock_i2s0_rx(Source(mux_i2s0_rx),
                           Control(Clock_gate_i2s0_rx),
                           Divider_i2s(Clock_divider_i2s0_m, Clock_divider_i2s0_n,
                                       Clock_divider_i2s0_d, Clock_divider_i2s0_n_auto,
                                       Clock_divider_i2s0_d_auto)),

             clock_i2s0_tx(Source(mux_i2s0_tx),
                           Control(Clock_gate_i2s0_tx),
                           Divider_i2s(Clock_divider_i2s1_m, Clock_divider_i2s1_n,
                                       Clock_divider_i2s1_d, Clock_divider_i2s1_n_auto,
                                       Clock_divider_i2s1_d_auto));

const double x1600_pll_intermediate_min = 600000000,
             x1600_pll_intermediate_max = 2400000000;

static Pll   clock_pll_A(Source(mux_external),
                         Control_pll(Pll_enable_A, Pll_stable_A, Pll_bypass_A),
                         Divider_pll(Pll_multiplier_A, Pll_input_division_A,
                                     Pll_output_division0_A, Pll_output_division1_A,
                                     x1600_pll_intermediate_min, x1600_pll_intermediate_max)),

             clock_pll_E(Source(mux_external),
                         Control_pll(Pll_enable_E, Pll_stable_E, Pll_bypass_E),
                         Divider_pll(Pll_multiplier_E, Pll_input_division_E,
                                     Pll_output_division0_E, Pll_output_division1_E,
                                     x1600_pll_intermediate_min, x1600_pll_intermediate_max)),

             clock_pll_M(Source(mux_external),
                         Control_pll(Pll_enable_M, Pll_stable_M, Pll_bypass_M),
                         Divider_pll(Pll_multiplier_M, Pll_input_division_M,
                                     Pll_output_division0_M, Pll_output_division1_M,
                                     x1600_pll_intermediate_min, x1600_pll_intermediate_max));



// Clock register.

static Clock_base *clocks[Clock_identifier_count] = {
  &clock_aic,
  &clock_none,            // Clock_aic_bitclk
  &clock_none,            // Clock_aic_pclk
  &clock_can0,
  &clock_can1,
  &clock_cdbus,
  &clock_cim,
  &clock_cpu,
  &clock_ddr,
  &clock_dma,
  &clock_none,            // Clock_emac
  &clock_external,
  &clock_hclock0,
  &clock_hclock2,
  &clock_hclock2_pclock,
  &clock_none,            // Clock_hdmi
  &clock_i2c0,
  &clock_i2c1,
  &clock_none,            // Clock_i2c2
  &clock_none,            // Clock_i2c3
  &clock_none,            // Clock_i2c4
  &clock_i2s0,            // supplies i2s0_rx
  &clock_i2s0_rx,
  &clock_i2s0_tx,
  &clock_i2s1,            // supplies i2s0_tx
  &clock_none,            // Clock_i2s1_rx
  &clock_none,            // Clock_i2s1_tx
  &clock_none,            // Clock_kbc
  &clock_none,            // Clock_lcd
  &clock_lcd_pixel,
  &clock_none,            // Clock_lcd_pixel1
  &clock_mac,
  &clock_main,
  &clock_mipi_csi,
  &clock_none,            // Clock_msc
  &clock_msc0,
  &clock_msc1,
  &clock_none,            // Clock_msc2
  &clock_none,            // Clock_none
  &clock_otg0,
  &clock_none,            // Clock_otg1
  &clock_pclock,
  &clock_pll_A,
  &clock_pll_E,
  &clock_pll_M,
  &clock_none,            // Clock_pll_V
  &clock_pwm0,
  &clock_none,            // Clock_pwm1
  &clock_none,            // Clock_rtc
  &clock_none,            // Clock_scc
  &clock_sfc,
  &clock_none,            // Clock_ssi
  &clock_ssi0,
  &clock_none,            // Clock_ssi1
  &clock_none,            // Clock_ssi2
  &clock_timer,
  &clock_uart0,
  &clock_uart1,
  &clock_uart2,
  &clock_uart3,
  &clock_none,            // Clock_udc
  &clock_none,            // Clock_uhc
  &clock_none,            // Clock_uprt
};



// Peripheral abstraction.

Cpm_x1600_chip::Cpm_x1600_chip(l4_addr_t addr)
: Cpm_chip(addr, clocks)
{
}



// C language interface functions.

void
*x1600_cpm_init(l4_addr_t cpm_base)
{
  return (void *) new Cpm_x1600_chip(cpm_base);
}

const char *
x1600_cpm_clock_type(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->clock_type(clock);
}

int
x1600_cpm_have_clock(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->have_clock(clock);
}

void
x1600_cpm_start_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_x1600_chip *>(cpm)->start_clock(clock);
}

void
x1600_cpm_stop_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_x1600_chip *>(cpm)->stop_clock(clock);
}

int
x1600_cpm_get_parameters(void *cpm, enum Clock_identifiers clock, uint32_t parameters[])
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_parameters(clock, parameters);
}

int
x1600_cpm_set_parameters(void *cpm, enum Clock_identifiers clock, int num_parameters, uint32_t parameters[])
{
  return static_cast<Cpm_x1600_chip *>(cpm)->set_parameters(clock, num_parameters, parameters);
}

uint8_t
x1600_cpm_get_source(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_source(clock);
}

void
x1600_cpm_set_source(void *cpm, enum Clock_identifiers clock, uint8_t source)
{
  static_cast<Cpm_x1600_chip *>(cpm)->set_source(clock, source);
}

enum Clock_identifiers
x1600_cpm_get_source_clock(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_source_clock(clock);
}

void
x1600_cpm_set_source_clock(void *cpm, enum Clock_identifiers clock, enum Clock_identifiers source)
{
  static_cast<Cpm_x1600_chip *>(cpm)->set_source_clock(clock, source);
}

uint32_t
x1600_cpm_get_source_frequency(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_source_frequency(clock);
}

uint32_t
x1600_cpm_get_frequency(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_frequency(clock);
}

int
x1600_cpm_set_frequency(void *cpm, enum Clock_identifiers clock, uint32_t frequency)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->set_frequency(clock, frequency);
}
