/*
 * DMA support for the X1600.
 *
 * Copyright (C) 2021, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/sys/types.h>
#include <stdint.h>



/* Enumerated types for various transfer parameters. */

enum Dma_x1600_request_type
{
  Dma_request_auto            = 8,
  Dma_request_can0_out        = 10,
  Dma_request_can0_in         = 11,
  Dma_request_can1_out        = 10,
  Dma_request_can1_in         = 11,
  Dma_request_uart3_out       = 14,
  Dma_request_uart3_in        = 15,
  Dma_request_uart2_out       = 16,
  Dma_request_uart2_in        = 17,
  Dma_request_uart1_out       = 18,
  Dma_request_uart1_in        = 19,
  Dma_request_uart0_out       = 20,
  Dma_request_uart0_in        = 21,
  Dma_request_ssi0_send_empty = 22,
  Dma_request_ssi0_recv_full  = 23,
  Dma_request_i2c0_send_empty = 36,
  Dma_request_i2c0_recv_full  = 37,
  Dma_request_i2c1_send_empty = 38,
  Dma_request_i2c1_recv_full  = 39,
  Dma_request_ssi_slv_out     = 42,
  Dma_request_ssi_slv_in      = 43,
  Dma_request_msc0_send_empty = 52,
  Dma_request_msc0_recv_full  = 53,
  Dma_request_msc1_send_empty = 54,
  Dma_request_msc1_recv_full  = 55,
  Dma_request_sadc_in         = 56,
  Dma_request_aic_loop_out    = 61,
  Dma_request_aic_out         = 62,
  Dma_request_aic_in          = 63,
};



#ifdef __cplusplus

#include <l4/devices/cpm-x1600.h>
#include <l4/devices/hw_mmio_register_block.h>

// Forward declaration.

class Dma_x1600_chip;



// DMA channel.

class Dma_x1600_channel
{
private:
  Hw::Register_block<32> _regs;
  Dma_x1600_chip *_chip;
  uint8_t _channel;
  l4_cap_idx_t _irq = L4_INVALID_CAP;

public:
  Dma_x1600_channel(Dma_x1600_chip *chip, uint8_t channel, l4_addr_t start, l4_cap_idx_t irq);

  unsigned int transfer(uint32_t source, uint32_t destination,
                        unsigned int count,
                        bool source_increment, bool destination_increment,
                        uint8_t source_width, uint8_t destination_width,
                        uint8_t transfer_unit_size,
                        enum Dma_x1600_request_type type=Dma_request_auto);

  unsigned int wait();

protected:
  // Transfer property configuration.

  uint32_t encode_req_detect_int_length(uint8_t units);

  uint32_t encode_source_port_width(uint8_t width);

  uint32_t encode_destination_port_width(uint8_t width);

  uint32_t encode_transfer_unit_size(uint8_t size);

  // Transaction control.

  void ack_irq();

  bool completed();

  bool error();

  bool halted();

  bool wait_for_irq();

  bool wait_for_irq(unsigned int timeout);
};

// DMA device control.

class Dma_x1600_chip
{
private:
  Hw::Register_block<32> _regs;
  l4_addr_t _start, _end;
  Cpm_x1600_chip *_cpm;

public:
  Dma_x1600_chip(l4_addr_t start, l4_addr_t end, Cpm_x1600_chip *cpm);

  void disable();

  void enable();

  Dma_x1600_channel *get_channel(uint8_t channel, l4_cap_idx_t irq);

  // Transaction control.

  void ack_irq(uint8_t channel);

  bool error();

  bool halted();

  bool have_interrupt(uint8_t channel);
};

#endif /* __cplusplus */



/* C language interface. */

EXTERN_C_BEGIN

void *x1600_dma_init(l4_addr_t start, l4_addr_t end, void *cpm);

void x1600_dma_disable(void *dma_chip);

void x1600_dma_enable(void *dma_chip);

void *x1600_dma_get_channel(void *dma, uint8_t channel, l4_cap_idx_t irq);

unsigned int x1600_dma_transfer(void *dma_channel,
                                uint32_t source, uint32_t destination,
                                unsigned int count,
                                int source_increment, int destination_increment,
                                uint8_t source_width, uint8_t destination_width,
                                uint8_t transfer_unit_size,
                                enum Dma_x1600_request_type type);

unsigned int x1600_dma_wait(void *dma_channel);

EXTERN_C_END
